/*
 * Copyright (c) 2010, AaYJFG - XBMC community
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the XBMC community nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL AaYJFG BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <fcntl.h>
#include <signal.h>
#include <libusb-1.0/libusb.h>

#include "xbmcclient.h"

#define msg(x...) do { fprintf(stdout, x); } while (0)
#define err(x...) do { fprintf(stderr, x); } while (0)

#define PCHK(x...)											\
	do {												\
		int __err = (x);					  				\
		if (__err == -1) {									\
			err("*** POSIX ERROR *** in %s:%d [%s]: ", __FILE__, __LINE__, #x);		\
			perror("");									\
			abort();									\
		}											\
	} while (0)

#define UCHK(x...)											\
	do {												\
		int __err = (x);					  				\
		if (__err < 0)										\
			err("*** USB ERROR *** in %s:%d [%s]: %d\n", __FILE__, __LINE__, #x, __err);	\
	} while (0)


static CXBMCClient xbmc;
static int disconnected = 0;


static void xbmc_action (const char *action)
{
	msg("send action '%s'\n", action);
	xbmc.SendACTION(action);
}

static void xbmc_action_button (const char *btn)
{
	msg("send action '%s' (button)\n", btn);
	xbmc.SendACTION(btn, ACTION_BUTTON);
}

static void xbmc_key (const char *key)
{
	msg("send key '%s'\n", key);
	xbmc.SendButton(key, "R1", BTN_NO_REPEAT);
}

static void xbmc_key_kb (const char *key)
{
	msg("send key '%s'\n", key);
	xbmc.SendButton(key, "KB", BTN_NO_REPEAT);
}

static void emit_left_click (void)
{
	msg("send left click\n");
	xbmc_action_button("select");
}

static void emit_right_click (void)
{
	msg("send right click\n");
	xbmc_action_button("contextmenu");
}

static void emit_mouse (int dx, int dy)
{
	if (dx != 0 || dy != 0) {
		const int maxcoord = 65535;
		static int x = maxcoord / 2;
		static int y = maxcoord / 2;
		x += dx * 64;
		y += dy * 64;
		if (x < 0)
			x = 0;
		if (y < 0)
			y = 0;
		if (x > maxcoord)
			x = maxcoord;
		if (y > maxcoord)
			y = maxcoord;
		msg("mouse dx %d, dy %d -- x %d y %d\n", dx, dy, x, y);
		xbmc.SendMOUSE(x, y);
	}
}

static void transfer0x81_cb (struct libusb_transfer *transfer)
{
	static const struct {
		char modifier;
		char code;
		void (*key_fn) (const char *key);
		const char *key;
	} translation_table [] = {
		{ 0x03, 0x17, xbmc_action, "XBMC.ActivateWindow(MyTV)" }, /* (yellow) TV */
		{ 0x01, 0x10, xbmc_action, "XBMC.ActivateWindow(MyMusic)" }, /* (blue) Music */
		{ 0x01, 0x0c, xbmc_action, "XBMC.ActivateWindow(MyPictures)" }, /* (green) Pictures */
		{ 0x01, 0x08, xbmc_action, "XBMC.ActivateWindow(MyVideos)" }, /* (red) Videos */
		{ 0x01, 0x12, xbmc_action_button, "AspectRatio" },
		{ 0x01, 0x0a, xbmc_action, "?" }, /**/
		{ 0x01, 0x17, xbmc_action, "?" }, /**/
		{ 0x03, 0x10, xbmc_action, "XBMC.EjectTray()" }, /* CD/DVD icon */
		{ 0x03, 0x05, xbmc_key, "reverse" }, /* Rewind */
		{ 0x03, 0x09, xbmc_key, "forward" }, /* FastForward */
		{ 0x01, 0x15, xbmc_action_button, "Queue" }, /* Record -> queue */
		{ 0x00, 0x2a, xbmc_key_kb, "escape" }, /* Back */
		{ 0x00, 0x28, xbmc_key_kb, "enter" }, /* ok -> enter */
		{ 0x00, 0x4f, xbmc_key, "right" },
		{ 0x00, 0x50, xbmc_key, "left" },
		{ 0x00, 0x51, xbmc_key, "down" },
		{ 0x00, 0x52, xbmc_key, "up" },
		{ 0x00, 0x4b, xbmc_key, "pageplus" }, /* ChannelUp -> PageUp*/
		{ 0x00, 0x4e, xbmc_key, "pageminus" }, /* ChannelDown -> PageUp */
		{ 0x0c, 0x28, xbmc_action, "XBMC.ActivateWindow(Home)" }, /* Start -> Home */
		{ 0x00, 0x59, xbmc_key, "one" },
		{ 0x00, 0x5a, xbmc_key, "two" },
		{ 0x00, 0x5b, xbmc_key, "three" },
		{ 0x00, 0x5c, xbmc_key, "four" },
		{ 0x00, 0x5d, xbmc_key, "five" },
		{ 0x00, 0x5e, xbmc_key, "six" },
		{ 0x00, 0x5f, xbmc_key, "seven" },
		{ 0x00, 0x60, xbmc_key, "eight" },
		{ 0x00, 0x61, xbmc_key, "nine" },
		{ 0x00, 0x62, xbmc_key, "zero" },
		{ 0x00, 0x55, xbmc_key, "star" },
		{ 0x04, 0x5d, xbmc_key, "hash" },
		{ 0x00, 0x29, xbmc_key, "clear" }, /* clear -> backspace */
		{ 0x04, 0x3d, xbmc_key, "back" }, /* close -> back */
		{ 0x04, 0x5b, NULL, NULL }, /* discard */
		{ 0x00, 0x53, NULL, NULL }, /* discard */
		{ 0x0c, 0x00, NULL, NULL }, /* discard */
		{ 0x08, 0x00, NULL, NULL }, /* discard */
		{ 0x04, 0x00, NULL, NULL }, /* discard */
		{ 0x03, 0x00, NULL, NULL }, /* discard */
		{ 0x01, 0x00, NULL, NULL }, /* discard */
	};
	static void (*repeat_fn) (const char *key) = NULL;
	static const char *repeat_key = NULL;
	if (transfer->status == LIBUSB_TRANSFER_NO_DEVICE) {
		disconnected = 1;
		return;
	}
	if (transfer->status == LIBUSB_TRANSFER_TIMED_OUT) {
		if (repeat_fn != NULL && repeat_key != NULL)
			repeat_fn(repeat_key);
		goto done;
	}
	if (transfer->actual_length == 8) {
		char m = transfer->buffer[0];
		char c = transfer->buffer[2];
		if (m == 0x00 && c == 0x00) {
			repeat_fn = NULL;
			repeat_key = NULL;
			goto done;
		}
		for (int i = 0; i < sizeof(translation_table)/sizeof(translation_table[0]); i++) {
			if (translation_table[i].modifier == m && translation_table[i].code == c) {
				if (translation_table[i].key_fn != NULL)
					translation_table[i].key_fn(translation_table[i].key);
				repeat_fn = translation_table[i].key_fn;
				repeat_key = translation_table[i].key;
				goto done;
			}
		}
	}
	if (transfer->actual_length > 0) {
		msg("*** unknown keycode: ");
		for (int i = 0; i < transfer->actual_length; i++)
			msg("%02x ", transfer->buffer[i] & 0xff);
		msg("\n");
	}
done:
	UCHK(libusb_submit_transfer(transfer));
}

static void transfer0x82_cb (struct libusb_transfer *transfer)
{
	static const struct {
		char code [5];
		void (*key_fn) (const char *key);
		const char *key;
	} translation_table [] = {
		{ { 0x03, 0x02, 0x55, 0x55, 0x55 }, xbmc_key, "power" },
		{ { 0x02, 0x02, 0x00, 0x00, 0x55 }, xbmc_key, "title" },
		{ { 0x02, 0x80, 0x00, 0x00, 0x55 }, xbmc_key, "skipminus" },
		{ { 0x02, 0x00, 0x02, 0x00, 0x55 }, xbmc_key, "skipplus" },
		{ { 0x02, 0x00, 0x00, 0x01, 0x55 }, xbmc_key, "stop" },
		{ { 0x02, 0x00, 0x00, 0x02, 0x55 }, xbmc_key, "pause" },
		{ { 0x02, 0x00, 0x01, 0x00, 0x55 }, xbmc_key, "mute" },
		{ { 0x02, 0x10, 0x00, 0x00, 0x55 }, xbmc_key, "volumeplus" },
		{ { 0x02, 0x00, 0x10, 0x00, 0x55 }, xbmc_key, "volumeminus" },
	};
	static void (*repeat_fn) (const char *key) = NULL;
	static const char *repeat_key = NULL;
	static int rclick_pending = 0;
	if (transfer->status == LIBUSB_TRANSFER_NO_DEVICE) {
		disconnected = 1;
		return;
	}
	if (transfer->status == LIBUSB_TRANSFER_TIMED_OUT) {
		if (rclick_pending == 1) {
			rclick_pending = 0;
			xbmc_key("info");
		}
		if (repeat_fn != NULL && repeat_key != NULL)
			repeat_fn(repeat_key);
		goto done;
	}
	if (transfer->actual_length == 5) {
		if (transfer->buffer[0] == 0x01) {
			if (transfer->buffer[1] == 0x00) {
				char dx = transfer->buffer[2];
				char dy = transfer->buffer[3];
				if (dx != 0 || dy != 0)
					emit_mouse(dx, dy);
				else if (rclick_pending > 0) {
					if (--rclick_pending == 0)
						emit_right_click();
				}
			} else if (transfer->buffer[1] == 0x01)
				emit_left_click();
			else if (transfer->buffer[1] == 0x02)
				rclick_pending = 2;

			repeat_fn = NULL;
			repeat_key = NULL;
			goto done;
		}

		const char keycode_up [5] = { 0x02, 0x00, 0x00, 0x00, 0x55 };
		const char keycode_up_power [5] = { 0x03, 0x00, 0x55, 0x55, 0x55 };
		if (memcmp(transfer->buffer, keycode_up, 5) == 0 ||
			memcmp(transfer->buffer, keycode_up_power, 5) == 0) {
			repeat_fn = NULL;
			repeat_key = NULL;
			goto done;
		}
		for (int i = 0; i < sizeof(translation_table)/sizeof(translation_table[0]); i++) {
			if (memcmp(translation_table[i].code, transfer->buffer, 5) == 0) {
				if (translation_table[i].key_fn != NULL)
					translation_table[i].key_fn(translation_table[i].key);
				repeat_fn = translation_table[i].key_fn;
				repeat_key = translation_table[i].key;
				goto done;
			}
		}
		msg("*** unknown keycode: ");
		for (int i = 0; i < transfer->actual_length; i++)
			msg("%02x ", transfer->buffer[i] & 0xff);
		msg("\n");
	}
done:
	UCHK(libusb_submit_transfer(transfer));
}

static void handle_exit(int sig)
{
	disconnected = 1;
}

int main (int argc, char **argv)
{
	struct sigaction sa;
	sigemptyset(&sa.sa_mask);
	sa.sa_handler = handle_exit;
	PCHK(sigaction(SIGINT, &sa, NULL));

	libusb_context *ctx;
	libusb_device_handle *dev;
	struct libusb_transfer *transfer0x81 = libusb_alloc_transfer(0);
	struct libusb_transfer *transfer0x82 = libusb_alloc_transfer(0);
	unsigned char buf0x81 [8];
	unsigned char buf0x82 [5];
	int i;

	UCHK(libusb_init(&ctx));

	if (!(dev = libusb_open_device_with_vid_pid(ctx, 0x05a4, 0x9881))) {
		err("%s: No HAMA MCE remote control found.\n", argv[0]);
		return -1;
	}

	msg("Connected HAMA MCE Remote\n");
	xbmc.SendHELO("HAMA MCE Remote", ICON_NONE);

	if (libusb_kernel_driver_active(dev, 0))
		UCHK(libusb_detach_kernel_driver(dev, 0));
	if (libusb_kernel_driver_active(dev, 1))
		UCHK(libusb_detach_kernel_driver(dev, 1));
	UCHK(libusb_claim_interface(dev, 0));
	UCHK(libusb_claim_interface(dev, 1));

	libusb_fill_interrupt_transfer(transfer0x81, dev, 0x81, buf0x81, sizeof(buf0x81), transfer0x81_cb, NULL, 215);
	UCHK(libusb_submit_transfer(transfer0x81));

	libusb_fill_interrupt_transfer(transfer0x82, dev, 0x82, buf0x82, sizeof(buf0x82), transfer0x82_cb, NULL, 200);
	UCHK(libusb_submit_transfer(transfer0x82));

	while (!disconnected)
		UCHK(libusb_handle_events(ctx));

	msg("Disconnected HAMA MCE Remote\n");
	xbmc.SendNOTIFICATION("Disconnected", "HAMA MCE Remote", ICON_NONE);
	// xbmc.SendBYE(); /* Not Implemented Yet */

	libusb_free_transfer(transfer0x81);
	libusb_free_transfer(transfer0x82);

	/* not needed, if we arrive here, the device has been disconnected.
	UCHK(libusb_release_interface(dev, 0));
	UCHK(libusb_release_interface(dev, 1));
	UCHK(libusb_attach_kernel_driver(dev, 0));
	UCHK(libusb_attach_kernel_driver(dev, 1));
	*/
 	libusb_close(dev);
	libusb_exit(ctx);

	return 0;
}

